

#include <util/delay.h>
#include "spi.h"




/********************************************
 * ENC28J60 memory and register definitions *
 ********************************************/


/* SPI commands */

#define RCR       0x00
#define RBM       0x3A
#define WCR       0x40
#define WBM       0x7A
#define BFS       0x80
#define BFC       0xA0
#define SRC       0xFF

/* Memory regions */

#define RXBUFST   0x0000
#define RXBUFND   0x17FF
#define TXBUFST   0x1800
#define TXBUFND   0x1FFF

/* Register flags */

#define DUMMY     0x20
#define BANK0     0x00
#define BANK1     0x40
#define BANK2     0x80
#define BANK3     0xC0
#define ADDRMASK  0x1F
#define BANKPOS   6

/* Common registers */

#define ESTAT     0x1D
#define   CLKRDY  0x01
#define ECON2     0x1E
#define   PKTDEC  0x40
#define ECON1     0x1F
#define   DMAST   0x20
#define   CSUMEN  0x10
#define   TXRTS   0x08
#define   RXEN    0x04
#define   BSEL1   0x02
#define   BSEL0   0x01

/* Bank 0 registers */

#define ERDPT     ( 0x00 | BANK0 )
#define EWRPT     ( 0x02 | BANK0 )
#define ETXST     ( 0x04 | BANK0 )
#define ETXND     ( 0x06 | BANK0 )
#define ERXST     ( 0x08 | BANK0 )
#define ERXND     ( 0x0A | BANK0 )
#define ERXRDPT   ( 0x0C | BANK0 )
#define ERXWRPT   ( 0x0E | BANK0 )
#define EDMAST    ( 0x10 | BANK0 )
#define EDMAND    ( 0x12 | BANK0 )
#define EDMACS    ( 0x16 | BANK0 )

/* Bank 1 registers */

#define ERXFCON   ( 0x18 | BANK1 )
#define   UCEN    0x80
#define EPKTCNT   ( 0x19 | BANK1 )

/* Bank 2 registers */

#define MACON1    ( 0x00 | BANK2 )
#define   MARXEN  0x01
#define MACON3    ( 0x02 | BANK2 )
#define   PADCFG2 0x80
#define   PADCFG1 0x40
#define   PADCFG0 0x20
#define   TXCRCEN 0x10
#define MICMD     ( 0x12 | BANK2 | DUMMY )
#define   MIIRD   0x01
#define MIREGADR  ( 0x14 | BANK2 | DUMMY )
#define MIWR      ( 0x16 | BANK2 | DUMMY )
#define MIRD      ( 0x18 | BANK2 | DUMMY )

/* Bank 3 registers */

#define MAADR5    ( 0x00 | BANK3 )
#define MAADR6    ( 0x01 | BANK3 )
#define MAADR3    ( 0x02 | BANK3 )
#define MAADR4    ( 0x03 | BANK3 )
#define MAADR1    ( 0x04 | BANK3 )
#define MAADR2    ( 0x05 | BANK3 )
#define MISTAT    ( 0x0A | BANK3 | DUMMY )
#define   BUSY    0x01

/* PHY registers */

#define PHCON2    0x10
#define   HDLDIS  0x0100
#define PHLCON    0x14
#define   LACFG3  0x0800
#define   LACFG2  0x0400
#define   LACFG1  0x0200
#define   LACFG0  0x0100
#define   LBCFG3  0x0080
#define   LBCFG2  0x0040
#define   LBCFG1  0x0020
#define   LBCFG0  0x0010
#define   LFRQ1   0x0008
#define   LFRQ0   0x0004
#define   STRCH   0x0002

/* Other stuff */

#define CRC32_LEN 4




/****************************
 * Read and write registers *
 ****************************/


/*
 * Select proper register bank according to the given address.
 */
void
enc28j60_bankSelect(uint8_t addr)
{
	static uint8_t currentBank = 0xFF;

	if (currentBank != (addr >> BANKPOS)) {
		currentBank = (addr >> BANKPOS);
		spi_select();
		spi_write(BFC | ECON1);
		spi_write(BSEL1 | BSEL0);
		spi_deselect();
		spi_select();
		spi_write(BFS | ECON1);
		spi_write(currentBank);
		spi_deselect();
	}
}


/*
 * Read a regular register.
 */
uint8_t
enc28j60_regReadByte(uint8_t addr)
{
	uint8_t data;

	enc28j60_bankSelect(addr);

	spi_select();
	spi_write(RCR | (addr & ADDRMASK));
	if (addr & DUMMY) {
		spi_write(0);
	}
	data = spi_read();
	spi_deselect();

	return data;
}


/*
 * Write a regular register.
 */
void
enc28j60_regWriteByte(uint8_t addr, uint8_t data)
{
	enc28j60_bankSelect(addr);

	spi_select();
	spi_write(WCR | (addr & ADDRMASK));
	spi_write(data);
	spi_deselect();
}


/*
 * Set single bits of a regular register.
 */
void
enc28j60_regSetBits(uint8_t addr, uint8_t data)
{
	enc28j60_bankSelect(addr);

	spi_select();
	spi_write(BFS | (addr & ADDRMASK));
	spi_write(data);
	spi_deselect();
}


/*
 * Clear single bits of a regular register.
 */
void
enc28j60_regClearBits(uint8_t addr, uint8_t data)
{
	enc28j60_bankSelect(addr);

	spi_select();
	spi_write(BFC | (addr & ADDRMASK));
	spi_write(data);
	spi_deselect();
}


/*
 * Perform a reset.
 */
void
enc28j60_reset(void)
{
	spi_select();
	spi_write(SRC);
	spi_deselect();
}


/*
 * Read a word from a regular register pair.
 */
uint16_t
enc28j60_regReadWord(uint8_t addr)
{
	return (enc28j60_regReadByte(addr + 0) << 0) |
	       (enc28j60_regReadByte(addr + 1) << 8);
}


/*
 * Write a word to a regular register pair.
 */
void
enc28j60_regWriteWord(uint8_t addr, uint16_t data)
{
	enc28j60_regWriteByte(addr + 0, data >> 0);
	enc28j60_regWriteByte(addr + 1, data >> 8);
}


/*
 * Read a word from a PHY register.
 */
uint16_t
enc28j60_phyReadWord(uint8_t addr)
{
	enc28j60_regWriteByte(MIREGADR, addr);
	enc28j60_regSetBits(MICMD, MIIRD);
	while (enc28j60_regReadByte(MISTAT) & BUSY);
	enc28j60_regClearBits(MICMD, MIIRD);

	return (enc28j60_regReadByte(MIRD + 0) << 0) |
	       (enc28j60_regReadByte(MIRD + 1) << 8);
}


/*
 * Write a word to a PHY register
 */
void
enc28j60_phyWriteWord(uint8_t addr, uint16_t data)
{
	while (enc28j60_regReadByte(MISTAT) & BUSY);
	enc28j60_regWriteByte(MIREGADR, addr);
	enc28j60_regWriteByte(MIWR + 0, data >> 0);
	enc28j60_regWriteByte(MIWR + 1, data >> 8);
}


/**********************
 * Read and write RAM *
 **********************/


/*
 * Read len bytes from RAM into data. ERDPT must be set before.
 */
void
enc28j60_bufRead(uint16_t len, uint8_t* data)
{
	spi_select();
	spi_write(RBM);
	while (len--) {
		*data++ = spi_read();
	}
	spi_deselect();
}


/*
 * Read a word from RAM. ERDPT must be set before.
 */
uint16_t
enc28j60_bufReadWord(void)
{
	uint16_t data;

	spi_select();
	spi_write(RBM);
	data  = spi_read() << 0;
	data |= spi_read() << 8;
	spi_deselect();

	return data;
}


/*
 * Write len bytes from data into RAM. EWRPT must be set before.
 */
void
enc28j60_bufWrite(uint16_t len, uint8_t* data)
{
	spi_select();
	spi_write(WBM);
	while (len--) {
		spi_write(*data++);
	}
	spi_deselect();
}


/*
 * Write a byte into RAM. EWRPT must be set before.
 */
void
enc28j60_bufWriteByte(uint8_t data)
{
	spi_select();
	spi_write(WBM);
	spi_write(data);
	spi_deselect();
}


/*
 * Initialization
 */
void
enc28j60_init(void)
{
	/* Reset ENC28J60 */
	_delay_ms(50);
	enc28j60_reset();
	_delay_ms(50);
	enc28j60_reset();
	_delay_ms(50);

	/* Setup status LEDs. */
	/* Orange = RX & TX, Green = link state */
	enc28j60_phyWriteWord(PHLCON, 0x0740 | STRCH);

	/* No loopback */
	enc28j60_phyWriteWord(PHCON2, HDLDIS);

	/* Setup RX & TX RAM boundaries */
	enc28j60_regWriteWord(ERXST, RXBUFST);
	enc28j60_regWriteWord(ERXND, RXBUFND);
	enc28j60_regWriteWord(ERXRDPT, RXBUFST);
	enc28j60_regWriteWord(ETXST, TXBUFST);

	/* Enable promiscuous mode */
	enc28j60_regWriteByte(ERXFCON, 0);

	/* Enable packet reception */
	enc28j60_regSetBits(MACON1, MARXEN);
	enc28j60_regSetBits(ECON1, RXEN);

	/* Active automatic CRC and padding generation */
	enc28j60_regWriteByte(MACON3, TXCRCEN | PADCFG2 | PADCFG1 | PADCFG0);
}


/*
 * Setup MAC address
 */
void
enc28j60_setMac(uint8_t* mac)
{
	enc28j60_regWriteByte(MAADR1, *mac++);
	enc28j60_regWriteByte(MAADR2, *mac++);
	enc28j60_regWriteByte(MAADR3, *mac++);
	enc28j60_regWriteByte(MAADR4, *mac++);
	enc28j60_regWriteByte(MAADR5, *mac++);
	enc28j60_regWriteByte(MAADR6, *mac++);
}


/*
 * Read the next available packet up to maxLen bytes to data
 * and store it's length. If no packet is available or it's length
 * exceeds maxLen, len is set to 0.
 */
void
enc28j60_readPacket(uint8_t* data, uint16_t* len, uint16_t maxLen)
{
	uint16_t nextPacketPointer;
	
	if (enc28j60_regReadByte(EPKTCNT) > 0) {

		enc28j60_regWriteWord(ERDPT, enc28j60_regReadWord(ERXRDPT));

		nextPacketPointer = enc28j60_bufReadWord();
		*len = enc28j60_bufReadWord() - CRC32_LEN;
		enc28j60_bufReadWord();

		if (*len <= maxLen) {
			enc28j60_bufRead(*len, data);
		} else {
			*len = 0;
		}

		enc28j60_regWriteWord(ERXRDPT, nextPacketPointer);
		enc28j60_regSetBits(ECON2, PKTDEC);

	} else {

		*len = 0;

	}
}


/*
 * Calculate checksum over data in RAM from startAdr to endAdr
 * and place the result at checksumAdr.
 */
void
enc28j60_calcChecksum(uint16_t startAdr, uint16_t endAdr, uint16_t checksumAdr)
{
	uint16_t checksum;

	/* Clear old checksum in RAM. */
	enc28j60_regWriteWord(EWRPT, TXBUFST + 1 + checksumAdr);
	enc28j60_bufWriteByte(0);
	enc28j60_bufWriteByte(0);

	/* Let ENC28J60 calculate the checksum for the specified RAM range. */
	enc28j60_regWriteWord(EDMAST, TXBUFST + 1 + startAdr);
	enc28j60_regWriteWord(EDMAND, TXBUFST + 1 + endAdr);
	enc28j60_regSetBits(ECON1, DMAST | CSUMEN);
	while (enc28j60_regReadByte(ECON1) & DMAST);
	checksum = enc28j60_regReadWord(EDMACS);

	/* Store calculated checksum at specified position in RAM. */
	enc28j60_regWriteWord(EWRPT, TXBUFST + 1 + checksumAdr);
	enc28j60_bufWriteByte(checksum >> 8);
	enc28j60_bufWriteByte(checksum >> 0);
}


/*
 * Copy len bytes from data into the transmit buffer in RAM.
 */
void
enc28j60_writePacket(uint8_t* data, uint16_t len)
{
	enc28j60_regWriteWord(EWRPT, TXBUFST);
	enc28j60_bufWriteByte(0);
	enc28j60_bufWrite(len, data);
	enc28j60_regWriteWord(ETXND, enc28j60_regReadWord(EWRPT) - 1);
}


/*
 * Initiate transmission of trasmit buffer.
 */
void
enc28j60_sendPacket(void)
{
	enc28j60_regSetBits(ECON1, TXRTS);
}
