

#include <avr/io.h>
#include <util/delay.h>
#include "debug.h"




/*
 * Soft-SPI interface definition
 */

#define RFM12_HW_PORT 		PORTD
#define RFM12_HW_PIN  		PIND
#define RFM12_HW_DDR  		DDRD

#define RFM12_HW_SDO		PD2
#define RFM12_HW_SEL		PD3
#define RFM12_HW_IRQ		PD4
#define RFM12_HW_SCK		PD5
#define RFM12_HW_DAT		PD6
#define RFM12_HW_SDI		PD7


/*
 * RFM12 command and bit definitions
 */

#define RFM12_CMD_STAT			0x0000
#define   RFM12_STAT_FFIT		0x8000
#define RFM12_CMD_SETCONF		0x8000
#define RFM12_CMD_PWRMANAGE 	0x8200
#define   RFM12_PWRMANAGE_ER	0x0080
#define   RFM12_PWRMANAGE_EBB	0x0040
#define   RFM12_PWRMANAGE_ET	0x0020
#define   RFM12_PWRMANAGE_ES	0x0010
#define   RFM12_PWRMANAGE_EX	0x0008
#define   RFM12_PWRMANAGE_EB	0x0004
#define   RFM12_PWRMANAGE_EW	0x0002
#define   RFM12_PWRMANAGE_DC	0x0001
#define RFM12_CMD_READFIFO		0xB000
#define RFM12_CMD_WRITEFIFO		0xB800


/*
 * Low level communication routines
 */

#define RFM12_SELECT 	{ RFM12_HW_PORT &= ~(1 << RFM12_HW_SEL); }
#define RFM12_DESELECT 	{ RFM12_HW_PORT |= (1 << RFM12_HW_SEL); }
#define RFM12_READ 		( (RFM12_HW_PIN & (1 << RFM12_HW_SDO)) != 0)
#define RFM12_WRITE_0 	{ RFM12_HW_PORT &= ~(1 << RFM12_HW_SDI); RFM12_HW_PORT |= (1 << RFM12_HW_SCK); RFM12_HW_PORT &= ~(1 << RFM12_HW_SCK); }
#define RFM12_WRITE_1 	{ RFM12_HW_PORT |= (1 << RFM12_HW_SDI); RFM12_HW_PORT |= (1 << RFM12_HW_SCK); RFM12_HW_PORT &= ~(1 << RFM12_HW_SCK); }
#define RFM12_BUSY		( (RFM12_HW_PIN & (1 << RFM12_HW_IRQ)) != 0 )
#define RFM12_IRQ		( (RFM12_HW_PIN & (1 << RFM12_HW_IRQ)) == 0 )




void rfm12_pollData(void);
uint16_t rfm12_sendCommand(uint16_t cmd);
void rfm12_initTx(void);
void rfm12_initRx(void);




static uint8_t* PacketBuf = 0;
static uint16_t PacketBufSize = 0;
static uint16_t PacketBufLength = 0;

static uint8_t Receiving = 0;




uint16_t
rfm12_sendCommand(uint16_t out)
{
	uint8_t i;
	uint16_t in = 0;

	RFM12_SELECT

	for (i=0; i<16; i++) {

		in <<= 1;
		if (RFM12_READ)
			in |= 0x0001;

		if (out & 0x8000)
			RFM12_WRITE_1
		else
			RFM12_WRITE_0
		out <<= 1;

	}

	RFM12_DESELECT

	return in;
}


void
rfm12_init(uint8_t* buf, uint16_t bufSize)
{
	PacketBuf = buf;
	PacketBufSize = bufSize;

	RFM12_HW_DDR &= ~((1 << RFM12_HW_SDO) | (1 << RFM12_HW_IRQ) | (1<< RFM12_HW_DAT));
	RFM12_HW_DDR |= ((1 << RFM12_HW_SEL) | (1 << RFM12_HW_SCK) | (1 << RFM12_HW_SDI));
	RFM12_HW_PORT |= ((1 << RFM12_HW_SEL) | (1 << RFM12_HW_DAT));
	RFM12_HW_PORT &= ~((1 << RFM12_HW_SCK));

	rfm12_sendCommand(0x80D8); // EL,EF,433band,12.5pF
	rfm12_sendCommand(0xA640); // 434MHz
	rfm12_sendCommand(0xC602); // 114 kbps ?!
	rfm12_sendCommand(0x94A0); // VDI,FAST,134kHz,0dBm,-103dBm
	rfm12_sendCommand(0xC2AC); // AL,!ml,DIG,DQD4
	rfm12_sendCommand(0xC4F7); // Offset independent from VDI, max deviation +3/-4 fres, FI, OE, EN
	rfm12_sendCommand(0x9850); // !mp,90kHz,MAX OUT
	rfm12_sendCommand(0xE000); // NOT USE
	rfm12_sendCommand(0xC800); // NOT USE
	rfm12_sendCommand(0xC400); // 1.66MHz,2.2V

	rfm12_initRx();
}


void
rfm12_initTx(void)
{
	rfm12_sendCommand(0x8201);// !er,!ebb,!et,!es,!ex,!eb,!ew,DC
	rfm12_sendCommand(0xCA81);// FIFO8,SYNC,!ff,DR
}


void
rfm12_initRx(void)
{
	rfm12_sendCommand(0x82D9); // ER,EBB,!et,ES,EX,!eb,!ew,DC
	rfm12_sendCommand(0xCA81); // FIFO8,SYNC,!ff,DR
	rfm12_sendCommand(0xCA83); // FIFO8,SYNC,FF,DR
}


void
rfm12_main(void)
{
	rfm12_pollData();
}


void
rfm12_pollData(void)
{
	uint16_t stat;
	uint8_t data;
	static uint16_t len = 0;
	static uint16_t pos = 0;

	stat = rfm12_sendCommand(RFM12_CMD_STAT);

	if (stat & RFM12_STAT_FFIT) {  /* There's a received byte waiting in the FIFO */

		Receiving = 1;

		data = rfm12_sendCommand(RFM12_CMD_READFIFO) & 0xFF;

		pos++;

		if (pos <= 2) {
			/* The first 2 bytes contain the packet length */
			len = (len << 8) | data;
			if (len > PacketBufSize) {
				/* Abort if packet doesn't fit into the packet buffer */
				len = 0;
				pos = 0;
				rfm12_initRx();
				Receiving = 0;
			}
		} else {
			/* Further bytes are written into the packet buffer */
			PacketBuf[pos - 3] = data;
			if (pos - 3 == len) {
				/* Packet reception completed */
				PacketBufLength = len;
				len = 0;
				pos = 0;
				rfm12_initRx();
				Receiving = 0;
			}
		}

	}
}


uint8_t
rfm12_isReceiving(void)
{
	return Receiving;
}


void
rfm12_readPacket(uint8_t* data, uint16_t* len, uint16_t maxLen)
{
	*len = PacketBufLength;
	PacketBufLength = 0;
}


void
rfm12_writePacket(uint8_t* data, uint16_t len)
{
	rfm12_initTx();

	rfm12_sendCommand(RFM12_CMD_STAT); // preload FIFO
	rfm12_sendCommand(RFM12_CMD_WRITEFIFO | 0xAA);
	rfm12_sendCommand(RFM12_CMD_STAT);
	rfm12_sendCommand(RFM12_CMD_WRITEFIFO | 0xAA);

	rfm12_sendCommand(0x8239); // enable transmitter

	while (RFM12_BUSY);        // Send sync word
	rfm12_sendCommand(RFM12_CMD_STAT);
	rfm12_sendCommand(RFM12_CMD_WRITEFIFO | 0x2D);
	while (RFM12_BUSY);
	rfm12_sendCommand(RFM12_CMD_STAT);
	rfm12_sendCommand(RFM12_CMD_WRITEFIFO | 0xD4);

	while (RFM12_BUSY);        // Send length
	rfm12_sendCommand(RFM12_CMD_STAT);
	rfm12_sendCommand(RFM12_CMD_WRITEFIFO + ((len & 0xFF00) >> 8));
	while (RFM12_BUSY);
	rfm12_sendCommand(RFM12_CMD_STAT);
	rfm12_sendCommand(RFM12_CMD_WRITEFIFO + ((len & 0x00FF) >> 0));

	while (len-- > 0) {
		while (RFM12_BUSY);
		rfm12_sendCommand(RFM12_CMD_STAT);
		rfm12_sendCommand(RFM12_CMD_WRITEFIFO + *data++);
	}

	while (RFM12_BUSY);        // Send dummy byte
	rfm12_sendCommand(RFM12_CMD_STAT);
	rfm12_sendCommand(RFM12_CMD_WRITEFIFO);

	while (RFM12_BUSY);
	rfm12_sendCommand(0x8219); // disable transmitter

	rfm12_initRx();
}
